/*
 * Copyright ©2023 Travis McGaha.  All rights reserved.  Permission is
 * hereby granted to students registered for University of Pennsylvania
 * CIT 5950 for use solely during Spring Semester 2023 for purposes of
 * the course.  No other use, copying, distribution, or modification
 * is permitted without prior written consent. Copyrights for
 * third-party components of this work must be honored.  Instructors
 * interested in reusing these course materials should contact the
 * author.
 */


#include "./Page.hpp"
#include "./PageTable.hpp"
#include "./catch.hpp"

using std::fstream;
using std::ios_base;
using simplevm::Page;
using simplevm::PageTable;

static constexpr const char* kSimpleFileName = "./test_files/simple.bytes";
static constexpr const char* kSimpleCopyFileName = "./test_files/simple_copy.bytes";
static constexpr const char* kComplexFileName = "./test_files/complex.bytes";
static constexpr const char* kComplexCopyFileName = "./test_files/complex_copy.bytes";

void setup_test_files() {
  fstream s(kSimpleFileName, ios_base::in);
  fstream copy;
  copy.open(kSimpleCopyFileName, ios_base::in | ios_base::out | ios_base::trunc);
  char bytes[Page::PAGE_SIZE];
  s.read(bytes, Page::PAGE_SIZE);
  while (!s.eof()) {
    copy.write(bytes, Page::PAGE_SIZE);
    s.read(bytes, Page::PAGE_SIZE);
  }

  copy.close();

  fstream c(kComplexFileName, ios_base::in);
  copy.open(kComplexCopyFileName, ios_base::in | ios_base::out | ios_base::trunc);
  c.read(bytes, Page::PAGE_SIZE);
  while (!c.eof()) {
    copy.write(bytes, Page::PAGE_SIZE);
    c.read(bytes, Page::PAGE_SIZE);
  }
}
  
TEST_CASE("ctor", "[Test_PageTable]") {
  setup_test_files();
  PageTable *pt = new PageTable(kSimpleCopyFileName, 2);
  for(uint32_t i = 0; i < 32U; i++) {
    REQUIRE_FALSE(pt->page_available(i));
  }
  REQUIRE(2U == pt->capacity());
  REQUIRE(0U == pt->loaded_pages()); 
  delete pt;
}

TEST_CASE("get_page_simple", "[Test_PageTable]") {
  setup_test_files();
  PageTable *pt = new PageTable(kSimpleCopyFileName, 2);
  REQUIRE(0U == pt->get_page(0).pno());
  REQUIRE(1U == pt->loaded_pages()); 
  REQUIRE_FALSE(pt->page_available(1U));
  REQUIRE_FALSE(pt->page_available(2U));
  REQUIRE_FALSE(pt->page_available(3U));

  REQUIRE(1U == pt->get_page(Page::PAGE_SIZE).pno());
  REQUIRE(pt->page_available(0U));
  REQUIRE(pt->page_available(1U));
  REQUIRE_FALSE(pt->page_available(2U));
  REQUIRE_FALSE(pt->page_available(3U));
  REQUIRE(2U == pt->capacity());
  REQUIRE(2U == pt->loaded_pages()); 

  REQUIRE(3U == pt->get_page(3 * Page::PAGE_SIZE + 3872).pno());
  REQUIRE(pt->page_available(1U));
  REQUIRE(pt->page_available(3U));
  REQUIRE_FALSE(pt->page_available(0U));
  REQUIRE_FALSE(pt->page_available(2U));
  REQUIRE(2U == pt->capacity());
  REQUIRE(2U == pt->loaded_pages()); 

  REQUIRE(1U == pt->get_page(6612).pno());
  REQUIRE(pt->page_available(1U));
  REQUIRE(pt->page_available(3U));
  REQUIRE_FALSE(pt->page_available(0U));
  REQUIRE_FALSE(pt->page_available(2U));
  REQUIRE(2U == pt->capacity());
  REQUIRE(2U == pt->loaded_pages()); 

  REQUIRE(2U == pt->get_page(10223).pno());
  REQUIRE(pt->page_available(1U));
  REQUIRE(pt->page_available(2U));
  REQUIRE_FALSE(pt->page_available(0U));
  REQUIRE_FALSE(pt->page_available(3U));
  REQUIRE(2U == pt->capacity());
  REQUIRE(2U == pt->loaded_pages()); 

  delete pt;
}

TEST_CASE("discard_page", "[Test_PageTable]") {
  setup_test_files();
  PageTable *pt = new PageTable(kSimpleCopyFileName, 3U);

  REQUIRE(0U == pt->get_page(0U).pno());
  REQUIRE(1U == pt->loaded_pages()); 
  REQUIRE_FALSE(pt->page_available(1U));
  REQUIRE_FALSE(pt->page_available(2U));
  REQUIRE_FALSE(pt->page_available(3U));

  // discarded pages not loaded should do nothing
  pt->discard_page(1U);
  pt->discard_page(2U);
  pt->discard_page(3U);
  REQUIRE(1U == pt->loaded_pages()); 
  REQUIRE(3U == pt->capacity());
  REQUIRE(pt->page_available(0U));
  REQUIRE_FALSE(pt->page_available(1U));
  REQUIRE_FALSE(pt->page_available(2U));
  REQUIRE_FALSE(pt->page_available(3U));

  // discard a page that is not the LRU
  REQUIRE(0U == pt->get_page(0U).pno());
  REQUIRE(1U == pt->get_page(1U * Page::PAGE_SIZE).pno());
  REQUIRE(2U == pt->loaded_pages()); 
  REQUIRE(3U == pt->capacity());

  pt->discard_page(1U);
  
  REQUIRE(1U == pt->loaded_pages()); 
  REQUIRE(3U == pt->capacity());
  REQUIRE(pt->page_available(0U));
  REQUIRE_FALSE(pt->page_available(1U));
  REQUIRE_FALSE(pt->page_available(2U));
  REQUIRE_FALSE(pt->page_available(3U));

  // Discard a page that is the LRU
  REQUIRE(0U == pt->get_page(0U).pno());
  REQUIRE(2U == pt->get_page(2U * Page::PAGE_SIZE + 13).pno());
  REQUIRE(1U == pt->get_page(2U * Page::PAGE_SIZE - 1).pno());

  pt->discard_page(0U);

  REQUIRE(2U == pt->loaded_pages()); 
  REQUIRE(3U == pt->capacity());
  REQUIRE_FALSE(pt->page_available(0U));
  REQUIRE(pt->page_available(1U));
  REQUIRE(pt->page_available(2U));
  REQUIRE_FALSE(pt->page_available(3U));

  // Clear the remaining pages
  pt->discard_page(2U);
  pt->discard_page(1U);
  REQUIRE(0U == pt->loaded_pages()); 
  REQUIRE(3U == pt->capacity());
  REQUIRE_FALSE(pt->page_available(0U));
  REQUIRE_FALSE(pt->page_available(1U));
  REQUIRE_FALSE(pt->page_available(2U));
  REQUIRE_FALSE(pt->page_available(3U));

  delete pt;
}

TEST_CASE("evict_page", "[Test_PageTable]") {
  setup_test_files();
  PageTable *pt = new PageTable(kComplexCopyFileName, 4U);
  REQUIRE(0U == pt->get_page(0U).pno());
  REQUIRE(1U == pt->loaded_pages()); 
  REQUIRE(4U == pt->capacity());

  REQUIRE(1U == pt->get_page(1U * Page::PAGE_SIZE).pno());
  REQUIRE(2U == pt->get_page(2U * Page::PAGE_SIZE).pno());
  REQUIRE(0U == pt->get_page(0U).pno());
  REQUIRE(3U == pt->get_page(3U * Page::PAGE_SIZE).pno());
  pt->evict_page();
  REQUIRE(3U == pt->loaded_pages()); 
  REQUIRE(4U == pt->capacity());
  REQUIRE(pt->page_available(0U));
  REQUIRE_FALSE(pt->page_available(1U));
  REQUIRE(pt->page_available(2U));
  REQUIRE(pt->page_available(3U));
  REQUIRE_FALSE(pt->page_available(4U));
  REQUIRE_FALSE(pt->page_available(5U));
  REQUIRE_FALSE(pt->page_available(6U));
  REQUIRE_FALSE(pt->page_available(7U));

  REQUIRE(4U == pt->get_page(4U * Page::PAGE_SIZE).pno());
  pt->evict_page();
  REQUIRE(3U == pt->loaded_pages()); 
  REQUIRE(4U == pt->capacity());
  REQUIRE(pt->page_available(0U));
  REQUIRE_FALSE(pt->page_available(1U));
  REQUIRE_FALSE(pt->page_available(2U));
  REQUIRE(pt->page_available(3U));
  REQUIRE(pt->page_available(4U));
  REQUIRE_FALSE(pt->page_available(5U));
  REQUIRE_FALSE(pt->page_available(6U));
  REQUIRE_FALSE(pt->page_available(7U));

  REQUIRE(5U == pt->get_page(5U * Page::PAGE_SIZE).pno());
  pt->evict_page();
  REQUIRE(3U == pt->loaded_pages()); 
  REQUIRE(4U == pt->capacity());
  REQUIRE_FALSE(pt->page_available(0U));
  REQUIRE_FALSE(pt->page_available(1U));
  REQUIRE_FALSE(pt->page_available(2U));
  REQUIRE(pt->page_available(3U));
  REQUIRE(pt->page_available(4U));
  REQUIRE(pt->page_available(5U));
  REQUIRE_FALSE(pt->page_available(6U));
  REQUIRE_FALSE(pt->page_available(7U));

  REQUIRE(0U == pt->get_page(0U * Page::PAGE_SIZE).pno());
  pt->evict_page();
  REQUIRE(3U == pt->loaded_pages()); 
  REQUIRE(4U == pt->capacity());
  REQUIRE(pt->page_available(0U));
  REQUIRE_FALSE(pt->page_available(1U));
  REQUIRE_FALSE(pt->page_available(2U));
  REQUIRE_FALSE(pt->page_available(3U));
  REQUIRE(pt->page_available(4U));
  REQUIRE(pt->page_available(5U));
  REQUIRE_FALSE(pt->page_available(6U));
  REQUIRE_FALSE(pt->page_available(7U));

  REQUIRE(1U == pt->get_page(1U * Page::PAGE_SIZE).pno());
  pt->evict_page();
  REQUIRE(3U == pt->loaded_pages()); 
  REQUIRE(4U == pt->capacity());
  REQUIRE(pt->page_available(0U));
  REQUIRE(pt->page_available(1U));
  REQUIRE_FALSE(pt->page_available(2U));
  REQUIRE_FALSE(pt->page_available(3U));
  REQUIRE_FALSE(pt->page_available(4U));
  REQUIRE(pt->page_available(5U));
  REQUIRE_FALSE(pt->page_available(6U));
  REQUIRE_FALSE(pt->page_available(7U));

  REQUIRE(2U == pt->get_page(2U * Page::PAGE_SIZE).pno());
  pt->evict_page();
  REQUIRE(3U == pt->loaded_pages()); 
  REQUIRE(4U == pt->capacity());
  REQUIRE(pt->page_available(0U));
  REQUIRE(pt->page_available(1U));
  REQUIRE(pt->page_available(2U));
  REQUIRE_FALSE(pt->page_available(3U));
  REQUIRE_FALSE(pt->page_available(4U));
  REQUIRE_FALSE(pt->page_available(5U));
  REQUIRE_FALSE(pt->page_available(6U));
  REQUIRE_FALSE(pt->page_available(7U));

  REQUIRE(3U == pt->get_page(3U * Page::PAGE_SIZE).pno());
  pt->evict_page();
  REQUIRE(3U == pt->loaded_pages()); 
  REQUIRE(4U == pt->capacity());
  REQUIRE_FALSE(pt->page_available(0U));
  REQUIRE(pt->page_available(1U));
  REQUIRE(pt->page_available(2U));
  REQUIRE(pt->page_available(3U));
  REQUIRE_FALSE(pt->page_available(4U));
  REQUIRE_FALSE(pt->page_available(5U));
  REQUIRE_FALSE(pt->page_available(6U));
  REQUIRE_FALSE(pt->page_available(7U));

  REQUIRE(4U == pt->get_page(4U * Page::PAGE_SIZE).pno());
  pt->evict_page();
  REQUIRE(3U == pt->loaded_pages()); 
  REQUIRE(4U == pt->capacity());
  REQUIRE_FALSE(pt->page_available(0U));
  REQUIRE_FALSE(pt->page_available(1U));
  REQUIRE(pt->page_available(2U));
  REQUIRE(pt->page_available(3U));
  REQUIRE(pt->page_available(4U));
  REQUIRE_FALSE(pt->page_available(5U));
  REQUIRE_FALSE(pt->page_available(6U));
  REQUIRE_FALSE(pt->page_available(7U));

  REQUIRE(7U == pt->get_page(7U * Page::PAGE_SIZE).pno());
  pt->evict_page();
  REQUIRE(3U == pt->loaded_pages()); 
  REQUIRE(4U == pt->capacity());
  REQUIRE_FALSE(pt->page_available(0U));
  REQUIRE_FALSE(pt->page_available(1U));
  REQUIRE_FALSE(pt->page_available(2U));
  REQUIRE(pt->page_available(3U));
  REQUIRE(pt->page_available(4U));
  REQUIRE_FALSE(pt->page_available(5U));
  REQUIRE_FALSE(pt->page_available(6U));
  REQUIRE(pt->page_available(7U));
  
  pt->evict_page();
  REQUIRE(2U == pt->loaded_pages()); 
  REQUIRE(4U == pt->capacity());
  REQUIRE_FALSE(pt->page_available(0U));
  REQUIRE_FALSE(pt->page_available(1U));
  REQUIRE_FALSE(pt->page_available(2U));
  REQUIRE_FALSE(pt->page_available(3U));
  REQUIRE(pt->page_available(4U));
  REQUIRE_FALSE(pt->page_available(5U));
  REQUIRE_FALSE(pt->page_available(6U));
  REQUIRE(pt->page_available(7U));

  pt->evict_page();
  REQUIRE(1U == pt->loaded_pages()); 
  REQUIRE(4U == pt->capacity());
  REQUIRE_FALSE(pt->page_available(0U));
  REQUIRE_FALSE(pt->page_available(1U));
  REQUIRE_FALSE(pt->page_available(2U));
  REQUIRE_FALSE(pt->page_available(3U));
  REQUIRE_FALSE(pt->page_available(4U));
  REQUIRE_FALSE(pt->page_available(5U));
  REQUIRE_FALSE(pt->page_available(6U));
  REQUIRE(pt->page_available(7U));

  pt->evict_page();
  REQUIRE(0U == pt->loaded_pages()); 
  REQUIRE(4U == pt->capacity());
  REQUIRE_FALSE(pt->page_available(0U));
  REQUIRE_FALSE(pt->page_available(1U));
  REQUIRE_FALSE(pt->page_available(2U));
  REQUIRE_FALSE(pt->page_available(3U));
  REQUIRE_FALSE(pt->page_available(4U));
  REQUIRE_FALSE(pt->page_available(5U));
  REQUIRE_FALSE(pt->page_available(6U));
  REQUIRE_FALSE(pt->page_available(7U));

  delete pt;
}

TEST_CASE("flush", "[Test_PageTable]") {
  setup_test_files();
  PageTable *pt = new PageTable(kComplexCopyFileName, 6U);
  uint32_t original5 = pt->get_page(5 * Page::PAGE_SIZE + 182).access<uint32_t>(5 * Page::PAGE_SIZE + 12);
  uint32_t original1 = pt->get_page(2 * Page::PAGE_SIZE - 37).access<uint32_t>(Page::PAGE_SIZE + 1);
  uint32_t original3 = pt->get_page(3 * Page::PAGE_SIZE + 2).access<uint32_t>(3 * Page::PAGE_SIZE + 1984);
  uint32_t original4 = pt->get_page(4 * Page::PAGE_SIZE + 90).access<uint32_t>(4 * Page::PAGE_SIZE + 2814);
  REQUIRE(4U == pt->loaded_pages()); 
  REQUIRE(6U == pt->capacity());

  pt->get_page(5 * Page::PAGE_SIZE + 182).store<uint32_t>(5 * Page::PAGE_SIZE + 12, original5 % 2);
  pt->get_page(2 * Page::PAGE_SIZE - 37).store<uint32_t>(Page::PAGE_SIZE + 1, original1 + 17);

  pt->flush_page(5U);

  // check that 5 was flushed
  uint32_t in_file;
  fstream reader(kComplexCopyFileName, ios_base::in);
  reader.seekg(5*Page::PAGE_SIZE + 12, ios_base::beg);
  reader.read(reinterpret_cast<char*>(&in_file), sizeof(in_file));

  REQUIRE((original5 % 2) == in_file);
  REQUIRE(4U == pt->loaded_pages()); 
  REQUIRE(6U == pt->capacity());
  REQUIRE(pt->page_available(1U));
  REQUIRE(pt->page_available(3U));
  REQUIRE(pt->page_available(4U));
  REQUIRE(pt->page_available(5U));

  // and that 1 was not flushed 
  reader.seekg(Page::PAGE_SIZE + 1, ios_base::beg);
  reader.read(reinterpret_cast<char*>(&in_file), sizeof(in_file));
  REQUIRE(original1 == in_file);

  // make sure that fluh_all_pages flushes every page
  pt->get_page(3 * Page::PAGE_SIZE).store<uint32_t>(3 * Page::PAGE_SIZE + 1984, original3 + 505);
  pt->flush_all_pages();
  REQUIRE(4U == pt->loaded_pages()); 
  REQUIRE(6U == pt->capacity());
  REQUIRE(pt->page_available(1U));
  REQUIRE(pt->page_available(3U));
  REQUIRE(pt->page_available(4U));
  REQUIRE(pt->page_available(5U));

  // check that 1 and 3 were flushed
  reader.seekg(Page::PAGE_SIZE + 1, ios_base::beg);
  reader.read(reinterpret_cast<char*>(&in_file), sizeof(in_file));
  REQUIRE((original1 + 17) == in_file);

  reader.seekg(3 * Page::PAGE_SIZE + 1984, ios_base::beg);
  reader.read(reinterpret_cast<char*>(&in_file), sizeof(in_file));
  REQUIRE((original3 + 505) == in_file);

  // write to page 4, then discard the page, checking to make sure discard flushed the page
  pt->get_page(4 * Page::PAGE_SIZE).store<uint32_t>(4 * Page::PAGE_SIZE + 2814, original4 + 21);
  pt->discard_page(4U);
  REQUIRE(3U == pt->loaded_pages()); 
  REQUIRE(6U == pt->capacity());
  REQUIRE(pt->page_available(1U));
  REQUIRE(pt->page_available(3U));
  REQUIRE(pt->page_available(5U));

  reader.seekg(4 * Page::PAGE_SIZE + 2814, ios_base::beg);
  reader.read(reinterpret_cast<char*>(&in_file), sizeof(in_file));
  REQUIRE(original4 + 21 == in_file);

  delete pt;
}

TEST_CASE("get_page_complex", "[Test_PageTable]") {
  setup_test_files();
  PageTable *pt = new PageTable(kComplexCopyFileName, 4U);
  REQUIRE(0U == pt->get_page(0U).pno());
  REQUIRE(1U == pt->loaded_pages()); 
  REQUIRE(4U == pt->capacity());

  REQUIRE(1U == pt->get_page(1U * Page::PAGE_SIZE).pno());
  REQUIRE(2U == pt->get_page(2U * Page::PAGE_SIZE).pno());
  REQUIRE(0U == pt->get_page(0U).pno());
  REQUIRE(3U == pt->get_page(3U * Page::PAGE_SIZE).pno());
  REQUIRE(4U == pt->loaded_pages()); 
  REQUIRE(4U == pt->capacity());
  REQUIRE(pt->page_available(0U));
  REQUIRE(pt->page_available(1U));
  REQUIRE(pt->page_available(2U));
  REQUIRE(pt->page_available(3U));
  REQUIRE_FALSE(pt->page_available(4U));
  REQUIRE_FALSE(pt->page_available(5U));
  REQUIRE_FALSE(pt->page_available(6U));
  REQUIRE_FALSE(pt->page_available(7U));

  REQUIRE(4U == pt->get_page(4U * Page::PAGE_SIZE).pno());
  REQUIRE(4U == pt->loaded_pages()); 
  REQUIRE(4U == pt->capacity());
  REQUIRE(pt->page_available(0U));
  REQUIRE_FALSE(pt->page_available(1U));
  REQUIRE(pt->page_available(2U));
  REQUIRE(pt->page_available(3U));
  REQUIRE(pt->page_available(4U));
  REQUIRE_FALSE(pt->page_available(5U));
  REQUIRE_FALSE(pt->page_available(6U));
  REQUIRE_FALSE(pt->page_available(7U));

  REQUIRE(5U == pt->get_page(5U * Page::PAGE_SIZE).pno());
  REQUIRE(4U == pt->loaded_pages()); 
  REQUIRE(4U == pt->capacity());
  REQUIRE(pt->page_available(0U));
  REQUIRE_FALSE(pt->page_available(1U));
  REQUIRE_FALSE(pt->page_available(2U));
  REQUIRE(pt->page_available(3U));
  REQUIRE(pt->page_available(4U));
  REQUIRE(pt->page_available(5U));
  REQUIRE_FALSE(pt->page_available(6U));
  REQUIRE_FALSE(pt->page_available(7U));

  REQUIRE(0U == pt->get_page(0U * Page::PAGE_SIZE).pno());
  REQUIRE(4U == pt->loaded_pages()); 
  REQUIRE(4U == pt->capacity());
  REQUIRE(pt->page_available(0U));
  REQUIRE_FALSE(pt->page_available(1U));
  REQUIRE_FALSE(pt->page_available(2U));
  REQUIRE(pt->page_available(3U));
  REQUIRE(pt->page_available(4U));
  REQUIRE(pt->page_available(5U));
  REQUIRE_FALSE(pt->page_available(6U));
  REQUIRE_FALSE(pt->page_available(7U));

  REQUIRE(1U == pt->get_page(1U * Page::PAGE_SIZE).pno());
  REQUIRE(4U == pt->loaded_pages()); 
  REQUIRE(4U == pt->capacity());
  REQUIRE(pt->page_available(0U));
  REQUIRE(pt->page_available(1U));
  REQUIRE_FALSE(pt->page_available(2U));
  REQUIRE_FALSE(pt->page_available(3U));
  REQUIRE(pt->page_available(4U));
  REQUIRE(pt->page_available(5U));
  REQUIRE_FALSE(pt->page_available(6U));
  REQUIRE_FALSE(pt->page_available(7U));

  REQUIRE(2U == pt->get_page(2U * Page::PAGE_SIZE).pno());
  REQUIRE(4U == pt->loaded_pages()); 
  REQUIRE(4U == pt->capacity());
  REQUIRE(pt->page_available(0U));
  REQUIRE(pt->page_available(1U));
  REQUIRE(pt->page_available(2U));
  REQUIRE_FALSE(pt->page_available(3U));
  REQUIRE_FALSE(pt->page_available(4U));
  REQUIRE(pt->page_available(5U));
  REQUIRE_FALSE(pt->page_available(6U));
  REQUIRE_FALSE(pt->page_available(7U));

  REQUIRE(3U == pt->get_page(3U * Page::PAGE_SIZE).pno());
  REQUIRE(4U == pt->loaded_pages()); 
  REQUIRE(4U == pt->capacity());
  REQUIRE(pt->page_available(0U));
  REQUIRE(pt->page_available(1U));
  REQUIRE(pt->page_available(2U));
  REQUIRE(pt->page_available(3U));
  REQUIRE_FALSE(pt->page_available(4U));
  REQUIRE_FALSE(pt->page_available(5U));
  REQUIRE_FALSE(pt->page_available(6U));
  REQUIRE_FALSE(pt->page_available(7U));

  REQUIRE(4U == pt->get_page(4U * Page::PAGE_SIZE).pno());
  REQUIRE(4U == pt->loaded_pages()); 
  REQUIRE(4U == pt->capacity());
  REQUIRE_FALSE(pt->page_available(0U));
  REQUIRE(pt->page_available(1U));
  REQUIRE(pt->page_available(2U));
  REQUIRE(pt->page_available(3U));
  REQUIRE(pt->page_available(4U));
  REQUIRE_FALSE(pt->page_available(5U));
  REQUIRE_FALSE(pt->page_available(6U));
  REQUIRE_FALSE(pt->page_available(7U));

  REQUIRE(7U == pt->get_page(7U * Page::PAGE_SIZE).pno());
  REQUIRE(4U == pt->loaded_pages()); 
  REQUIRE(4U == pt->capacity());
  REQUIRE_FALSE(pt->page_available(0U));
  REQUIRE_FALSE(pt->page_available(1U));
  REQUIRE(pt->page_available(2U));
  REQUIRE(pt->page_available(3U));
  REQUIRE(pt->page_available(4U));
  REQUIRE_FALSE(pt->page_available(5U));
  REQUIRE_FALSE(pt->page_available(6U));
  REQUIRE(pt->page_available(7U));
  
  delete pt;
}
